@@ -1,26 +1,19 @@
 //
-// "$Id: Fl_File_Browser.cxx,v 1.1.2.25 2003/01/30 21:41:41 easysw Exp $"
+// "$Id: Fl_File_Browser.cxx 12047 2016-10-19 05:55:03Z manolo $"
 //
 // Fl_File_Browser routines.
 //
-// Copyright 1999-2003 by Michael Sweet.
+// Copyright 1999-2010 by Michael Sweet.
 //
-// This library is free software; you can redistribute it and/or
-// modify it under the terms of the GNU Library General Public
-// License as published by the Free Software Foundation; either
-// version 2 of the License, or (at your option) any later version.
+// This library is free software. Distribution and use rights are outlined in
+// the file "COPYING" which should have been included with this file.  If this
+// file is missing or damaged, see the license at:
 //
-// This library is distributed in the hope that it will be useful,
-// but WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-// Library General Public License for more details.
+//     http://www.fltk.org/COPYING.php
 //
-// You should have received a copy of the GNU Library General Public
-// License along with this library; if not, write to the Free Software
-// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
-// USA.
+// Please report all bugs and problems on the following page:
 //
-// Please report all bugs and problems to "fltk-bugs@fltk.org".
+//     http://www.fltk.org/str.php
 //
 // Contents:
 //
@@ -40,6 +33,7 @@
 #include <FL/Fl_File_Browser.H>
 #include <FL/fl_draw.H>
 #include <FL/filename.H>
+#include <FL/Fl_Image.H>	// icon
 #include <stdio.h>
 #include <stdlib.h>
 #include "flstring.h"
@@ -62,14 +56,38 @@
 #  include <os2.h>
 #endif // __EMX__
 
-// CodeWarrior (__MWERKS__) gets its include paths confused, so we
-// temporarily disable this...
-#if defined(__APPLE__) && !defined(__MWERKS__)
+#if defined(__APPLE__)
 #  include <sys/param.h>
 #  include <sys/ucred.h>
 #  include <sys/mount.h>
-#endif // __APPLE__ && !__MWERKS__
+#endif // __APPLE__
 
+#if defined(_AIX)
+extern "C" {
+#  include <sys/types.h>
+#  include <sys/vmount.h>
+#  include <sys/mntctl.h>
+  // Older AIX versions don't expose this prototype
+  int
+  mntctl(int, int, char *);
+}
+#endif  // _AIX
+
+#if defined(__NetBSD__)
+extern "C" {
+#  include <sys/param.h>  // For '__NetBSD_Version__' definition
+#  if defined(__NetBSD_Version__) && (__NetBSD_Version__ >= 300000000)
+#    include <sys/types.h>
+#    include <sys/statvfs.h>
+#    if defined(HAVE_PTHREAD) && defined(HAVE_PTHREAD_H)
+#      include <pthread.h>
+#    endif  // HAVE_PTHREAD && HAVE_PTHREAD_H
+#    ifdef HAVE_PTHREAD
+  static pthread_mutex_t getvfsstat_mutex = PTHREAD_MUTEX_INITIALIZER;
+#    endif  // HAVE_PTHREAD/
+#  endif  // __NetBSD_Version__
+}
+#endif  // __NetBSD__
 
 //
 // FL_BLINE definition from "Fl_Browser.cxx"...
@@ -78,11 +96,17 @@
 #define SELECTED 1
 #define NOTDISPLAYED 2
 
+// TODO -- Warning: The definition of FL_BLINE here is a hack.
+//    Fl_File_Browser should not do this. PLEASE FIX.
+//    FL_BLINE should be private to Fl_Browser, and not re-defined here.
+//    For now, make sure this struct is precisely consistent with Fl_Browser.cxx.
+//
 struct FL_BLINE			// data is in a linked list of these
 {
   FL_BLINE	*prev;		// Previous item in list
   FL_BLINE	*next;		// Next item in list
   void		*data;		// Pointer to data (function)
+  Fl_Image      *icon;		// Pointer to optional icon
   short		length;		// sizeof(txt)-1, may be longer than string
   char		flags;		// selected, displayed
   char		txt[1];		// start of allocated array
@@ -165,13 +189,16 @@ Fl_File_Browser::item_width(void *p) const	// I - List item data
   const int	*columns;		// Columns
 
 
-  // Set the font and size...
-  fl_font(textfont(), textsize());
-
   // Scan for newlines...
   line    = (FL_BLINE *)p;
   columns = column_widths();
 
+  // Set the font and size...
+  if (line->txt[strlen(line->txt) - 1] == '/')
+    fl_font(textfont() | FL_BOLD, textsize());
+  else
+    fl_font(textfont(), textsize());
+
   if (strchr(line->txt, '\n') == NULL &&
       strchr(line->txt, column_char()) == NULL)
   {
@@ -324,9 +351,8 @@ Fl_File_Browser::item_draw(void *p,	// I - List item data
   else
     fl_color(fl_inactive(c));
 
-  for (t = line->txt, ptr = fragment; *t != '\0'; t ++)
-    if (*t == '\n')
-    {
+  for (t = line->txt, ptr = fragment; *t != '\0'; t ++) {
+    if (*t == '\n') {
       // Newline - nul terminate this fragment and draw it...
       *ptr = '\0';
 
@@ -338,21 +364,18 @@ Fl_File_Browser::item_draw(void *p,	// I - List item data
       width  = 0;
       Y      += fl_height();
       column = 0;
-    }
-    else if (*t == column_char())
-    {
+    } else if (*t == column_char()) {
       // Tab - nul terminate this fragment and draw it...
       *ptr = '\0';
 
       int cW = W - width; // Clip width...
 
-      if (columns)
-      {
-        // Try clipping inside this column...
-	for (i = 0; i < column && columns[i]; i ++);
+      if (columns) {
+	// Try clipping inside this column...
+	for (i = 0; i < column && columns[i]; i ++) { ; }
 
-        if (columns[i])
-          cW = columns[i];
+	if (columns[i])
+	  cW = columns[i];
       }
 
       fl_draw(fragment, X + width, Y, cW, fl_height(),
@@ -360,21 +383,18 @@ Fl_File_Browser::item_draw(void *p,	// I - List item data
 
       // Advance to the next column...
       column ++;
-      if (columns)
-      {
+      if (columns) {
 	for (i = 0, width = 0; i < column && columns[i]; i ++)
 	  width += columns[i];
       }
       else
-        width = column * (int)(fl_height() * 0.6 * 8.0);
-
+	width = column * (int)(fl_height() * 0.6 * 8.0);
       ptr = fragment;
     }
     else
       *ptr++ = *t;
-
-  if (ptr > fragment)
-  {
+  }
+  if (ptr > fragment) {
     // Nul terminate this fragment and draw it...
     *ptr = '\0';
 
@@ -421,8 +441,12 @@ Fl_File_Browser::load(const char     *directory,// I - Directory to load
 //  printf("Fl_File_Browser::load(\"%s\")\n", directory);
 
   clear();
+
   directory_ = directory;
 
+  if (!directory)
+    return (0);
+
   if (directory_[0] == '\0')
   {
     //
@@ -461,7 +485,7 @@ Fl_File_Browser::load(const char     *directory,// I - Directory to load
       {
         sprintf(filename, "%c:/", i);
 
-	if (i < 'C')
+	if (i < 'C') // see also: GetDriveType and GetVolumeInformation in WIN32
 	  add(filename, icon);
 	else
 	  add(filename, icon);
@@ -488,12 +512,16 @@ Fl_File_Browser::load(const char     *directory,// I - Directory to load
 
 	num_files ++;
       }
-#elif defined(__APPLE__) && !defined(__MWERKS__)
+#elif defined(__APPLE__)
     // MacOS X and Darwin use getfsstat() system call...
     int			numfs;	// Number of file systems
     struct statfs	*fs;	// Buffer for file system info
 
 
+    // We always have the root filesystem.
+    add("/", icon);
+
+    // Get the mounted filesystems...
     numfs = getfsstat(NULL, 0, MNT_NOWAIT);
     if (numfs > 0) {
       // We have file systems, get them...
@@ -502,11 +530,11 @@ Fl_File_Browser::load(const char     *directory,// I - Directory to load
 
       // Add filesystems to the list...
       for (i = 0; i < numfs; i ++) {
-        if (fs[i].f_mntonname[1]) {
+	// Ignore "/", "/dev", and "/.vol"...
+        if (fs[i].f_mntonname[1] && strcmp(fs[i].f_mntonname, "/dev") &&
+	    strcmp(fs[i].f_mntonname, "/.vol")) {
           snprintf(filename, sizeof(filename), "%s/", fs[i].f_mntonname);
           add(filename, icon);
-        } else {
-          add("/", icon);
         }
         num_files ++;
       }
@@ -514,27 +542,101 @@ Fl_File_Browser::load(const char     *directory,// I - Directory to load
       // Free the memory used for the file system info array...
       delete[] fs;
     }
+#elif defined(_AIX)
+    // AIX don't write the mounted filesystems to a file like '/etc/mnttab'.
+    // But reading the list of mounted filesystems from the kernel is possible:
+    // http://publib.boulder.ibm.com/infocenter/pseries/v5r3/topic/com.ibm.aix.basetechref/doc/basetrf1/mntctl.htm
+    int res = -1, len;
+    char *list = NULL, *name;
+    struct vmount *vp;
+
+    // We always have the root filesystem
+    add("/", icon);
+    // Get the required buffer size for the vmount structures
+    res = mntctl(MCTL_QUERY, sizeof(len), (char *) &len);
+    if (!res) {
+      // Allocate buffer ...
+      list = (char *) malloc((size_t) len);
+      if (NULL == list) {
+        res = -1;
+      } else {
+        // ... and read vmount structures from kernel
+        res = mntctl(MCTL_QUERY, len, list);
+        if (0 >= res) {
+          res = -1;
+        } else {
+          for (i = 0, vp = (struct vmount *) list; i < res; ++i) {
+            name = (char *) vp + vp->vmt_data[VMT_STUB].vmt_off;
+            strlcpy(filename, name, sizeof(filename));
+            // Skip the already added root filesystem
+            if (strcmp("/", filename) != 0) {
+              strlcat(filename, "/", sizeof(filename));
+              add(filename, icon);
+            }
+            vp = (struct vmount *) ((char *) vp + vp->vmt_length);
+          }
+        }
+      }
+    }
+    // Note: Executing 'free(NULL)' is allowed and simply do nothing
+    free((void *) list);
+#elif defined(__NetBSD__) && defined(__NetBSD_Version__) \
+      && (__NetBSD_Version__ >= 300000000)
+    // NetBSD don't write the mounted filesystems to a file like '/etc/mnttab'.
+    // Since NetBSD 3.0 the system call getvfsstat(2) has replaced getfsstat(2)
+    // that is used by getmntinfo(3):
+    // http://www.daemon-systems.org/man/getmntinfo.3.html
+    int res = -1;
+    struct statvfs *list;
+
+    // We always have the root filesystem
+    add("/", icon);
+#  ifdef HAVE_PTHREAD
+    // Lock mutex for thread safety
+    if (!pthread_mutex_lock(&getvfsstat_mutex)) {
+#  endif  // HAVE_PTHREAD
+      // Get list of statvfs structures
+      res = getmntinfo(&list, ST_WAIT);
+      if(0 < res) {
+        for (i = 0;  i < res; ++i) {
+          strlcpy(filename, list[i].f_mntonname, sizeof(filename));
+          // Skip the already added root filesystem
+          if (strcmp("/", filename) != 0) {
+            strlcat(filename, "/", sizeof(filename));
+            add(filename, icon);
+          }
+        }
+      } else {
+         res = -1;
+      }
+#  ifdef HAVE_PTHREAD
+      pthread_mutex_unlock(&getvfsstat_mutex);
+    }
+#  endif  // HAVE_PTHREAD
 #else
     //
     // UNIX code uses /etc/fstab or similar...
     //
     FILE	*mtab;		// /etc/mtab or /etc/mnttab file
-    char	line[1024];	// Input line
+    char	line[FL_PATH_MAX];	// Input line
 
     //
     // Open the file that contains a list of mounted filesystems...
     //
 
-    mtab = fopen("/etc/mnttab", "r");	// Fairly standard
+    mtab = fl_fopen("/etc/mnttab", "r");	// Fairly standard
     if (mtab == NULL)
-      mtab = fopen("/etc/mtab", "r");	// More standard
+      mtab = fl_fopen("/etc/mtab", "r");	// More standard
     if (mtab == NULL)
-      mtab = fopen("/etc/fstab", "r");	// Otherwise fallback to full list
+      mtab = fl_fopen("/etc/fstab", "r");	// Otherwise fallback to full list
     if (mtab == NULL)
-      mtab = fopen("/etc/vfstab", "r");	// Alternate full list file
+      mtab = fl_fopen("/etc/vfstab", "r");	// Alternate full list file
 
     if (mtab != NULL)
     {
+      // We always have the root filesystem
+      add("/", icon);
+      num_files ++;
       while (fgets(line, sizeof(line), mtab) != NULL)
       {
         if (line[0] == '#' || line[0] == '\n')
@@ -542,6 +644,8 @@ Fl_File_Browser::load(const char     *directory,// I - Directory to load
         if (sscanf(line, "%*s%4095s", filename) != 1)
 	  continue;
 
+        if (strcmp("/", filename) == 0) continue; // root was added before
+        // Add a trailing slash (except for the root filesystem)
         strlcat(filename, "/", sizeof(filename));
 
 //        printf("Fl_File_Browser::load() - adding \"%s\" to list...\n", filename);
@@ -550,8 +654,13 @@ Fl_File_Browser::load(const char     *directory,// I - Directory to load
       }
 
       fclose(mtab);
+    } else {
+      // Every Unix has a root filesystem '/'.
+      // This last stage fallback ensures that the user don't get an empty
+      // window after requesting filesystem list.
+      add("/", icon);
     }
-#endif // WIN32 || __EMX__
+#endif // WIN32 || __EMX__ || __APPLE__ || _AIX || ...
   }
   else
   {
@@ -564,7 +673,7 @@ Fl_File_Browser::load(const char     *directory,// I - Directory to load
 
 #if (defined(WIN32) && !defined(__CYGWIN__)) || defined(__EMX__)
     strlcpy(filename, directory_, sizeof(filename));
-    i = strlen(filename) - 1;
+    i = (int) (strlen(filename) - 1);
 
     if (i == 2 && filename[1] == ':' &&
         (filename[2] == '/' || filename[2] == '\\'))
@@ -580,38 +689,20 @@ Fl_File_Browser::load(const char     *directory,// I - Directory to load
     if (num_files <= 0)
       return (0);
 
-    for (i = 0, num_dirs = 0; i < num_files; i ++)
-    {
-#if 0
-      if (strcmp(files[i]->d_name, ".") != 0 &&
-          strcmp(files[i]->d_name, "..") != 0)
-#else
-      if (strcmp(files[i]->d_name, ".") != 0)
-#endif // 0
-      {
+    for (i = 0, num_dirs = 0; i < num_files; i ++) {
+      if (strcmp(files[i]->d_name, "./")) {
 	snprintf(filename, sizeof(filename), "%s/%s", directory_,
 	         files[i]->d_name);
 
-#if defined(WIN32) && !defined(__CYGWIN__)
-	if (files[i]->d_name[strlen(files[i]->d_name) - 1] == '/')
-	{
+        icon = Fl_File_Icon::find(filename);
+	if ((icon && icon->type() == Fl_File_Icon::DIRECTORY) ||
+	     _fl_filename_isdir_quick(filename)) {
           num_dirs ++;
-          insert(num_dirs, files[i]->d_name, Fl_File_Icon::find(filename));
+          insert(num_dirs, files[i]->d_name, icon);
+	} else if (filetype_ == FILES &&
+	           fl_filename_match(files[i]->d_name, pattern_)) {
+          add(files[i]->d_name, icon);
 	}
-#else
-	if (fl_filename_isdir(filename))
-	{
-	  char name[1024]; // Temporary directory name
-
-	  snprintf(name, sizeof(name), "%s/", files[i]->d_name);
-
-          num_dirs ++;
-          insert(num_dirs, name, Fl_File_Icon::find(filename));
-	}
-#endif // WIN32 && !__CYGWIN__
-	else if (filetype_ == FILES &&
-	         fl_filename_match(files[i]->d_name, pattern_))
-          add(files[i]->d_name, Fl_File_Icon::find(filename));
       }
 
       free(files[i]);
@@ -636,12 +727,9 @@ Fl_File_Browser::filter(const char *pattern)	// I - Pattern string
     pattern_ = pattern;
   else
     pattern_ = "*";
-
-  // Reload the current directory...
-  load(directory_);
 }
 
 
 //
-// End of "$Id: Fl_File_Browser.cxx,v 1.1.2.25 2003/01/30 21:41:41 easysw Exp $".
+// End of "$Id: Fl_File_Browser.cxx 12047 2016-10-19 05:55:03Z manolo $".
 //
