@@ -1,33 +1,47 @@
 //
-// "$Id: filename_match.cxx,v 1.5.2.3 2000/06/05 21:21:05 mike Exp $"
+// "$Id: filename_match.cxx 8864 2011-07-19 04:49:30Z greg.ercolano $"
 //
 // Pattern matching routines for the Fast Light Tool Kit (FLTK).
 //
-// Copyright 1998-2000 by Bill Spitzak and others.
+// Copyright 1998-2010 by Bill Spitzak and others.
 //
-// This library is free software; you can redistribute it and/or
-// modify it under the terms of the GNU Library General Public
-// License as published by the Free Software Foundation; either
-// version 2 of the License, or (at your option) any later version.
+// This library is free software. Distribution and use rights are outlined in
+// the file "COPYING" which should have been included with this file.  If this
+// file is missing or damaged, see the license at:
 //
-// This library is distributed in the hope that it will be useful,
-// but WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-// Library General Public License for more details.
+//     http://www.fltk.org/COPYING.php
 //
-// You should have received a copy of the GNU Library General Public
-// License along with this library; if not, write to the Free Software
-// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
-// USA.
+// Please report all bugs and problems on the following page:
 //
-// Please report all bugs and problems to "fltk-bugs@fltk.org".
+//     http://www.fltk.org/str.php
 //
 
 /* Adapted from Rich Salz. */
 #include <FL/filename.H>
 #include <ctype.h>
 
-int filename_match(const char *s, const char *p) {
+/**
+    Checks if a string \p s matches a pattern \p p. 
+    The following syntax is used for the pattern:
+    - * matches any sequence of 0 or more characters.
+    - ? matches any single character.
+    - [set] matches any character in the set. Set can contain any single characters, or a-z to represent a range. 
+      To match ] or - they must be the first characters. To match ^ or ! they must not be the first characters.
+    - [^set] or [!set] matches any character not in the set.
+    - {X|Y|Z} or {X,Y,Z} matches any one of the subexpressions literally.
+    - \\x quotes the character x so it has no special meaning.
+    - x all other characters must be matched exactly.
+
+    \b Include:
+    \code
+    #include <FL/filename.H>
+    \endcode
+
+    \param[in] s the string to check for a match
+    \param[in] p the string pattern 
+    \return non zero if the string matches the pattern
+*/
+int fl_filename_match(const char *s, const char *p) {
   int matched;
 
   for (;;) {
@@ -39,7 +53,7 @@ int filename_match(const char *s, const char *p) {
 
     case '*' :	// match 0-n of any characters
       if (!*p) return 1; // do trailing * quickly
-      while (!filename_match(s, p)) if (!*s++) return 0;
+      while (!fl_filename_match(s, p)) if (!*s++) return 0;
       return 1;
 
     case '[': {	// match one character in set of form [abc-d] or [^a-b]
@@ -63,7 +77,7 @@ int filename_match(const char *s, const char *p) {
 
     case '{' : // {pattern1|pattern2|pattern3}
     NEXTCASE:
-    if (filename_match(s,p)) return 1;
+    if (fl_filename_match(s,p)) return 1;
     for (matched = 0;;) {
       switch (*p++) {
       case '\\': if (*p) p++; break;
@@ -89,25 +103,17 @@ int filename_match(const char *s, const char *p) {
     case 0:	// end of pattern
       return !*s;
 
-#if defined(WIN32) && !defined(__CYGWIN__)
     case '\\':	// quote next character
       if (*p) p++;
-      if (*s++ != *(p-1)) return 0;
-      break;
+      /* FALLTHROUGH */
     default:
       if (tolower(*s) != tolower(*(p-1))) return 0;
       s++;
-#else
-    case '\\':	// quote next character
-      if (*p) p++;
-    default  :
-      if (*s++ != *(p-1)) return 0;
       break;
-#endif
     }
   }
 }
 
 //
-// End of "$Id: filename_match.cxx,v 1.5.2.3 2000/06/05 21:21:05 mike Exp $".
+// End of "$Id: filename_match.cxx 8864 2011-07-19 04:49:30Z greg.ercolano $".
 //
